import logging
import threading
from typing import Type, Optional, Union, Any

from kontur.clusterconfig.default_config import DEFAULT_AUTO_UPDATE_SETTINGS_PERIOD, DEFAULT_ZONE, \
    ZONE_PARAM_DEFAULT_NAME, HOST_PARAM_DEFAULT_NAME, DEFAULT_BASE_URI, DEFAULT_LOCAL_DIRECTORY, \
    LOCAL_DIR_PARAM_DEFAULT_NAME, DEFAULT_PORT
from kontur.clusterconfig.helpers import Empty, load_cc_settings
from kontur.clusterconfig.loaders import AbstractSettingsLoader, ClusterSettingsLoader, LocalSettingsLoader


logger = logging.getLogger(__name__)


class ClusterConfigClient(object):
    """
    Базовый класс для клиента.

    """
    empty_instance = Empty()
    _cluster_settings_loader_class: Type[AbstractSettingsLoader] = ClusterSettingsLoader
    _local_settings_loader_class: Type[AbstractSettingsLoader] = LocalSettingsLoader
    _timer_class = threading.Timer

    def __init__(
            self,
            zone: Optional[str] = None,
            base_uri: Optional[str] = None,
            local_directory: Optional[str] = None,
            auto_update_settings_period: Optional[Union[float, int]] = DEFAULT_AUTO_UPDATE_SETTINGS_PERIOD
    ):
        """
        Клиент для ClusterConfig.

        :param Optional[str] zone: Зона для ClusterConfig
        :param Optional[str] base_uri: Базовый URL для ClusterConfig
        :param Optional[str] local_directory: Локальная директория, для поиска на диске
        :param Optional[Union[int, float]] auto_update_settings_period:

        """
        # Загрузка настроек с данными по ClusterConfig из файла
        _settings = load_cc_settings() if local_directory is None else load_cc_settings(local_directory)

        # Теперь сверяем откуда брать настройки.
        if zone is None:
            zone = _settings.get_param_or_default(ZONE_PARAM_DEFAULT_NAME, DEFAULT_ZONE)
        if base_uri is None:
            base_uri = _settings.get_param_or_default(HOST_PARAM_DEFAULT_NAME, DEFAULT_BASE_URI)
        if local_directory is None:
            local_directory = _settings.get_param_or_default(LOCAL_DIR_PARAM_DEFAULT_NAME, DEFAULT_LOCAL_DIRECTORY)

        # Теперь обрабатываем Host ClusterConfig.
        delim_idx = base_uri.find(':')
        if delim_idx != -1:
            _host = base_uri[0:delim_idx]
            _port = int(base_uri[(delim_idx + 1):])
        else:
            _port, _host = DEFAULT_PORT, base_uri

        base_uri = f'http://{_host}:{_port}'

        if base_uri[-1] == '/':
            base_uri = base_uri[0:-1]

        # Инициализируем переменные
        self._auto_update_settings_period = auto_update_settings_period
        self._current_settings = None
        self._locker = threading.Lock()
        self._cluster_settings_loader = self._cluster_settings_loader_class(base_uri, zone)
        self._local_settings_loader = self._local_settings_loader_class(local_directory)

        # Запускаем таймер для автообновления
        self._update_settings_timer = None
        if self._auto_update_settings_period:
            logger.info(f'Starting update settings timer, '
                        f'auto update period={self._auto_update_settings_period}')
            self._start_timer()

    def __del__(self):
        self.close()

    def close(self):
        """
        Останавливаем таймер.

        """
        if self._update_settings_timer and self._update_settings_timer.is_alive():
            logger.info('Stopping update settings timer')
            self._update_settings_timer.cancel()

    def get_all_settings(self) -> dict:
        """
        Возвращаем все конфиги.
        Если нужно, то предварительно их скачиваем.

        :return: Словарь со всеми настройками.
        :rtype: dict

        """
        with self._locker:
            if not self._current_settings:
                self._update_settings()
            return self._current_settings.copy()

    def get_settings_by_key(self, key: str, default: Any = None) -> Any:
        """
        Поиск значени в конфиге по ключу.

        :param str key: Ключ, по которому ищем.
        :param Any default: Дефолтное значение. Значение будет возвращено, если по ключу ничего не нашлось.
                            Если не передавать аргумент, вместо возвращаемого None будет выбрасываться исключение.

        :return: Найденное значение.
        :rtype: Any

        :raise KeyError: Если в конфиге не нашлось переменной по ключу `key=`.

        """
        default = default or ClusterConfigClient.empty_instance
        settings = self.get_all_settings()

        key = key.lower()
        value = settings.get(key, settings.get(key + '/', self.empty_instance))

        if isinstance(value, Empty):
            if isinstance(default, Empty):
                raise KeyError(f'Key {key} not found in clusterconfig')
            return default
        return value

    def _update_settings_by_timer(self):
        """
        Обновление конфигов по таймеру. Хэндлер
        После обновления, запускает таймер заново.

        """
        try:
            with self._locker:
                self._update_settings()
        except BaseException as e:
            logging.warning(f'Error updating settings: {e}')
        finally:
            if self._auto_update_settings_period:
                self._start_timer()

    def _start_timer(self):
        """
        Создает и запускает таймер по обновлению конфигов.

        """
        self._update_settings_timer = self._timer_class(self._auto_update_settings_period,
                                                        self._update_settings_by_timer)
        self._update_settings_timer.daemon = True
        self._update_settings_timer.start()

    def _update_settings(self):
        """
        Обновление конфигов из кластера и из файлов.

        """
        logger.debug('Start updating settings')

        cluster_settings = self._cluster_settings_loader.load()
        local_settings = self._local_settings_loader.load()

        self._current_settings = self._merge_settings(local_settings, cluster_settings)

        logger.debug('Finish updating settings')

    def _merge_settings(self, local_settings: dict, cluster_settings: dict) -> dict:
        """
        Мердж двух словарей с конфигами.
        Конфиги из кластера, можгут быть перезаписаны конфигами из файлов.

        :param dict local_settings: Словарь с конфигами из файлов.
        :param dict cluster_settings: Словарь с конфигами из кластера.

        :return: Словарь с помердженными конфигами.
        :rtype: dict

        """
        settings = cluster_settings
        for key, value in local_settings.items():
            settings[key] = value
        return settings
