"""
Загрузчики конфигов.

"""
import os
import logging
from abc import ABCMeta, abstractmethod
from typing import List

import requests
import six

from kontur.clusterconfig.errors import ClusterConfigClientError
from kontur.clusterconfig.helpers import find_directory_up_the_tree
from kontur.clusterconfig.settings import Settings

logger = logging.getLogger(__name__)


class AbstractSettingsLoader(six.with_metaclass(ABCMeta)):
    """
    Базовый класс загрузчика конфигов.

    """
    @abstractmethod
    def load(self) -> dict:
        """
        Загрузка переменных.

        """
        pass


class LocalSettingsLoader(AbstractSettingsLoader):
    """
    Загрузчик конфигов из файлов.

    """
    def __init__(self, local_directory: str):
        """
        Загрузчик конфигов из файлов.

        :param str local_directory: Путь до локальной директории, в которой лежат файлы с конфигами.

        """
        self._current_settings = None
        self._last_modified_ts = None
        self._local_directory = local_directory

    def load(self) -> dict:
        """
        Загрузка переменных.

        :return: Словарь с готовыми переменными.
        :rtype: dict

        """
        logger.debug('Start loading local settings, last-modified=`{}`'.format(self._last_modified_ts))

        path = find_directory_up_the_tree(self._local_directory)

        if not os.path.exists(path):
            return {}

        files = self._get_setting_files(path)

        # Проверяем, обновились ли конфиги, и если обновились читаем заново
        last_modified_ts = self._get_last_modified_ts(path, files)
        if not self._last_modified_ts or last_modified_ts > self._last_modified_ts:
            self._current_settings = self._load_settings(path, files)
            self._last_modified_ts = last_modified_ts

        logger.debug('Finish loading local settings')

        return self._current_settings

    def _get_setting_files(self, path: str) -> List[str]:
        """
        Собираем все пути до всех файлов с конфигами.

        :param str path: Путь до папки, в которой будем собирать все файлы с конфигами.

        :return: Список путей до файлов с конфигами.
        :rtype: List[str]

        """
        files = []
        for (dirpath, dirnames, filenames) in os.walk(path):
            files.extend([os.path.join(dirpath, file) for file in filenames])
        return files

    def _get_last_modified_ts(self, path: str, files: List[str]) -> float:
        """
        Возвращает самое свежее время изменения, среди всех файлов в иерархии файлов конфига.

        :param str path: Путь до корня с файлами конфигов.
        :param List[str] files: Список путей до каждого файла с конфигом.

        :return: Самое позднее время изменения файла.
        :rtype: float

        """
        times = [os.path.getmtime(path)]
        times.extend([os.path.getmtime(file) for file in files])
        return max(times)

    def _load_settings(self, path: str, files: List[str]) -> dict:
        """
        Загрузка всех конфигов из всех файлов.

        :param str path: Путь до корневой папки с конфигами.
        :param List[str] files: Список путей до все файлов конфигов.

        :return: Загруженные настройки.
        :rtype: dict

        """
        logger.debug('Start loading local settings from disk')
        settings = {}

        for file in files:
            prefix = file[(len(path) + 1):].replace("\\", "/", -1)
            settings.update(Settings(file, prefix).get_all_params())

        logger.debug('Finish loading local settings from disk')
        return settings


class ClusterSettingsLoader(AbstractSettingsLoader):
    """
    Загрузчик конфигов их кластера.

    """
    def __init__(self, base_uri: str, zone: str):
        """
        Загрузчик конфигов их кластера.

        :param str base_uri: URL до кластера с конфигами.
        :param str zone: Зона в кластере с конфигами.

        """
        self._current_settings = {}
        self._last_modified_ts = None
        self._base_uri = base_uri
        self._zone = zone
        self._errors_count = 0

    def load(self) -> dict:
        """
        Загрузка переменных.

        :return: Словарь с готовыми переменными.
        :rtype: dict

        :raise ClientError: Если кластер ответил с ошибкой на запрос.

        """
        logger.debug('Start loading cluster settings, last-modified={}'.format(self._last_modified_ts))

        response = self._query_settings()
        if self._is_successful_response(response):
            self._errors_count = 0
            if response.status_code == requests.codes.ok:
                self._current_settings = response.json()
                if 'last-modified' in response.headers:
                    self._last_modified_ts = response.headers['last-modified']
            elif response.status_code == requests.codes.not_found:
                raise ClusterConfigClientError(f'Zone {self._zone} not found on clusterconfig {self._base_uri}')
            elif response.status_code != requests.codes.not_modified:
                raise ClusterConfigClientError(f'Unknown response {response} from clusterconfig {self._base_uri}')
        else:
            if response.status_code == 503:
                self._errors_count += 1
            if self._errors_count >= 2:
                self._errors_count = 0
                raise ClusterConfigClientError(f'Clusterconfig request to {self._base_uri} failed: {response}')

        logger.debug('Finish loading cluster settings')

        return self._current_settings

    def _query_settings(self) -> requests.Response:
        """
        Запрос на настройки из кластера.

        :return: Ответ кластера.
        :rtype: requests.Response

        """
        headers = None
        if self._last_modified_ts:
            headers = {'if-modified-since': self._last_modified_ts}

        url = '{}/{}'.format(self._base_uri, self._zone)

        return requests.get(url, headers=headers)

    def _is_successful_response(self, response: requests.Response) -> bool:
        """
        Успешный ли запрос?

        :param requests.Response response: Ответ кластера.

        :return: Результат проверки.
        :rtype: bool

        """
        return response.ok or response.status_code == requests.codes.not_found
