"""
Асинхронные парсеры строк конфига.

"""
from abc import ABC, abstractmethod
from typing import Iterable, Optional

import aiofiles

from kontur.clusterconfig.errors import ParseLocalConfigError
from kontur.clusterconfig.parsers.mixins import LinesParserMixin


class AsyncAbstractParser(ABC):

    @abstractmethod
    async def parse(self, filename: str, prefix: str) -> dict:
        """
        Метод для парсинга файла
        :param str filename: Путь до файла, в котором лежат переменные.
        :param str prefix: Префикс, который устанавливается перед key при сохранении распарсенного значения.

        :return: Распарсенные переменные.
        :rtype: dict
        """


class AsyncKeyValueParser(AsyncAbstractParser, LinesParserMixin):
    """Асинхронный Парсер `key=value` строк."""

    def __init__(self, separators: Optional[Iterable[str]] = None):
        self.separators = separators or self.separators

    async def parse(self, filename: str, prefix: str = '') -> dict:
        """
        Сам процесс парсинга файла.

        :param str filename: Путь до файла, в котором лежат переменные.
        :param str prefix: Префикс, который устанавливается перед key при сохранении распарсенного значения.

        :return: Распарсенные переменные.
        :rtype: dict

        """
        result = {}
        async with aiofiles.open(filename, 'rt') as f:
            try:
                async for line in f:
                    result = self._processing_line(line, result, prefix)
            except ValueError as e:
                raise ParseLocalConfigError(e)
        return result


class AsyncSimpleFormatParser(AsyncAbstractParser):
    """Асинхронный парсинг файла с одним значением."""

    async def parse(self, filename: str, prefix: str) -> dict:
        """
        Сам процесс парсинга файла.

        :param str filename: Путь до файла, в котором лежит значение.
        :param str prefix: Название, которое стоит присвоить переменной.

        :return: Распарсенная переменная {prefix: value}
        :rtype: dict

        """
        async with aiofiles.open(filename, 'r') as f:
            value = await f.read()
        return {prefix: value}
