"""
Парсеры строк конфига.

"""
from typing import Iterable, Optional
from abc import abstractmethod, ABC

from kontur.clusterconfig.errors import ParseLocalConfigError
from kontur.clusterconfig.parsers.mixins import LinesParserMixin


class AbstractParser(ABC):
    """
    Базовый абстрактный класс парсера переменных.

    """
    @abstractmethod
    def parse(self, filename: str, prefix: str) -> dict:
        """
        Сам процесс парсинга.

        :param str filename: Путь до файла, в котором лежат переменные.
        :param str prefix: Префикс, который устанавливается перед key при сохранении распарсенного значения.

        :return: Распарсенные переменные.
        :rtype: dict

        """
        pass


class KeyValueParser(AbstractParser, LinesParserMixin):
    """
    Парсер `key=value` строк.

    """
    def __init__(self, separators: Optional[Iterable[str]] = None):
        """
        Парсер `key=value` строк.

        :param Iterable[str] separators: Список возможных разделителей KEY от VALUE.

        """
        self.separators = separators or self.separators

    def parse(self, filename: str, prefix: str = '') -> dict:
        """
        Сам процесс парсинга файла.

        :param str filename: Путь до файла, в котором лежат переменные.
        :param str prefix: Префикс, который устанавливается перед key при сохранении распарсенного значения.

        :return: Распарсенные переменные.
        :rtype: dict

        """
        result = {}
        with open(filename, 'rt') as f:
            try:
                for line in f:
                    result = self._processing_line(line, result, prefix)
            except ValueError as e:
                raise ParseLocalConfigError(e)
        return result


class SimpleFormatParser(AbstractParser):
    """
    Парсинг файла с одним значением.

    """
    def parse(self, filename: str, prefix: str) -> dict:
        """
        Сам процесс парсинга файла.

        :param str filename: Путь до файла, в котором лежит значение.
        :param str prefix: Название, которое стоит присвоить переменной.

        :return: Распарсенная переменная {prefix: value}
        :rtype: dict

        """
        with open(filename, 'r') as f:
            value = f.read()
        return {prefix: value}
