"""
Работа с настройками.

"""
import os
from typing import Any, Type

from kontur.clusterconfig.errors import SettingsError, ParseLocalConfigError
from kontur.clusterconfig.parsers import AbstractParser, SimpleFormatParser, KeyValueParser


class Settings(object):
    """
    Класс настроек, который умеет парсить переменные из файлов.

    """
    _simple_format_parser_class: Type[AbstractParser] = SimpleFormatParser
    _key_value_parser_class: Type[AbstractParser] = KeyValueParser

    def __init__(self, filename: str, prefix: str = ''):
        """
        Класс настроек, который умеет парсить переменные из файлов.

        :param str filename: Название файла, который будем парсить.
        :param str prefix: Префикс, который устанавливается перед key при сохранении распарсенного значения.

        """
        self.filename = filename
        self.parser = self._init_parser(filename)
        self.prefix = prefix
        self.settings = None

    def _ensure_settings(self):
        """
        Загружаем настройки.

        """
        if self.settings is None:
            self._reread_from_file()

    def _reread_from_file(self):
        """
        Парсим настройки из файла.

        """
        try:
            self.settings = self.parser.parse(self.filename, self.prefix)
        except (FileNotFoundError, ParseLocalConfigError):
            self.settings = {}

    def _init_parser(self, filename: str) -> AbstractParser:
        """
        Инициализация парсера файла с переменными.

        :param str filename: Название файла, который будем парсить.

        :return: Готовый к работе парсер.
        :rtype: BaseParserClass

        """
        simple_format_file_extensions = ('.xml', '.xslt', '.json', '.cs')
        file_extension = self._get_file_extension(filename)

        if file_extension in simple_format_file_extensions:
            return self._simple_format_parser_class()
        return self._key_value_parser_class()

    def _get_file_extension(self, filename: str) -> str:
        """
        Возвращает расширение файла.

        :param str filename: Название файла, у которого мы хотим вытащить расширение.

        :return: Расширение файла.
        :rtype: str

        """
        _, file_extension = os.path.splitext(filename)
        return file_extension

    def has_param(self, key: str) -> bool:
        """
        Проверка наличия параметра в конфигах по названию.

        :param str key: Название переменной, наличие которой проверяем.

        :return: Результат проверки наличия.
        :rtype: bool

        """
        self._ensure_settings()
        return key in self.settings

    def get_param_or_default(self, key: str, default: Any) -> Any:
        """
        Возвращает переменную по названию.
        Если переменной нет, возвращает default значение.

        :param str key: Название переменной, которую хотим получить.
        :param object default: Дефолтное значение переменной.

        :return: Найденное значение переменной.
        :rtype: object

        """
        self._ensure_settings()
        if self.has_param(key):
            return self.settings[key]
        return default

    def get_param(self, key: str) -> Any:
        """
        Возвращает значение переменной.

        :param str key: Название переменной, которую нужно вернуть.

        :return: Значение переменной.
        :rtype: object

        :raise SettingsError: Если переменная не найдена.

        """
        self._ensure_settings()
        if self.has_param(key):
            return self.settings[key]
        raise SettingsError('Key {} not found in config file {}'.format(key, self.filename))

    def get_all_params(self) -> dict:
        """
        Возвращает все настройки.

        :return: Все доступные настройки.
        :rtype: dict

        """
        self._ensure_settings()
        return self.settings.copy()
