import asyncio
import json
import logging
from copy import deepcopy
from pathlib import Path
from kontur.clusterconfig import AsyncClusterConfigClient


class Config:
    _default_ = {
        "log_level": "info",
        "api_key": "",
        "listen": "127.0.0.1",
        "listen_port": 2480,
        "tcp_message_format": "lles",
        "tcp_connector_limit": 100,
        "tcp_connector_limit_per_host": 10,
        "reader_buffer": 1024*1024*32,
        "compress": False,
        "topology": "direct",
        "topology_direct": ["https://hercules.skbkontur.ru:6306"],
        "topology_clusterconfig_address": "clusterconfig:9000",
        "topology_clusterconfig_key": "topology/hercules/gate.prod/",
        "flush_timeout": 10,
        "send_timeout": 20,
        "keepalive_timeout": 60
    }

    def __new__(cls, loop):
        if not hasattr(cls, 'instance'):
            cls.instance = super(Config, cls).__new__(cls)
        return cls.instance

    def __init__(self, loop):
        self.loop = loop
        self.logger = logging.getLogger()
        self.queue = asyncio.Queue()
        raw_config = self._get_raw_config()

        for k, v in self._default_.items():
            if raw_config is not None and k in raw_config:
                self.__dict__[k] = raw_config[k]
            else:
                self.__dict__[k] = deepcopy(v)

        #Modifications
        self.log_level = self._convert_log_level(self.log_level)
        logging.basicConfig(level=self.log_level)
        self.logger = logging.getLogger()

        #Checks
        if self.tcp_message_format not in ("lles", "lels"):
            self.logger.error("Config. tcp_message_format must be one of 'lles' or 'lels'")
            raise ValueError
        if self.topology not in ("direct", "clusterconfig"):
            self.logger.error("Config. topology must be one of ('direct', 'clusterconfig')")
            raise ValueError
        if type(self.topology_direct) is not list or not all((type(i) == str for i in self.topology_direct)):
            self.logger.error("Config. topology_direct must be list of strings")
            raise ValueError

        if self.loop is not None:
            if self.topology == "direct":
                self.hercules_topology = self.topology_direct
            if self.topology == "clusterconfig":
                self._cc = AsyncClusterConfigClient(base_uri=self.topology_clusterconfig_address)
                self.loop.create_task(self.update_cc_topology())

    async def update_cc_topology(self):
        while True:
            value = await self._cc.get_settings_by_key(self.topology_clusterconfig_key)
            if type(value) is str:
                self.hercules_topology = [value, ]
            await asyncio.sleep(30)

    def get_config(self):
        c = {k: deepcopy(self.__dict__[k]) for k in self._default_.keys()}
        c["log_level"] = self._convert_log_level(c["log_level"])
        return c

    def _convert_log_level(self, log_level):
        levels = {"info": logging.INFO,
                  "warning": logging.WARNING,
                  "error": logging.ERROR,
                  "critical": logging.CRITICAL,
                  "debug": logging.DEBUG}
        if type(log_level) is str:
            return levels.get(log_level, logging.INFO)
        for k, v in levels.items():
            if v == log_level:
                return k

    def _get_raw_config(self):
        configs_path = (Path("config.json"), Path("/etc/hercules-sender.json"))
        for path in configs_path:
            if not path.is_file():
                continue
            try:
                with open(path) as f:
                    filecontent = f.read()
                    if len(filecontent) == 0:
                        continue
                    return json.loads(filecontent)
            except IOError as err:
                self.logger.error("Config. IO Error when reading config")
                raise err
        self.logger.error("Config. config.json not found! Used default config")
        return None


if __name__ == "__main__":
    config = Config(None)
    print(json.dumps(config.get_config()))
