import aiohttp
import asyncio
import signal
import sys
from random import choice

import lz4.frame
import uvloop

from hercules_sender.config import Config
from hercules_sender.reader.lels import tcp_reader as lels_reader
from hercules_sender.reader.lles import tcp_reader as lles_reader


class HerculesSender:
    def __init__(self, loop, config):
        self._loop = loop
        self._waiter = asyncio.Event()
        self._flush_future = self._loop.create_task(self.flush_loop())
        self._config = config
        self._count_of_events = 0
        self._lock_for_count = asyncio.Lock()
        self.logger = config.logger
        self.running = True
        self._tcp_connector = aiohttp.TCPConnector(
            limit=self._config.tcp_connector_limit,
            limit_per_host=self._config.tcp_connector_limit_per_host,
            keepalive_timeout=float(self._config.keepalive_timeout))

    async def flush_loop(self):
        while self.running:
            try:
                timeout = self._config.flush_timeout
                await asyncio.wait_for(self._waiter.wait(),
                                       timeout=timeout,
                                       loop=self._loop)
            except asyncio.TimeoutError:
                pass

            try:
                await self.flush()
            except Exception as ex:
                self.logger.error("Sender[flush_loop]. Exception: {}".format(ex))

            self._waiter.clear()

    async def flush(self):
        self.logger.debug("Sender[flush]. Queue flusher is awoke")
        max_content_length = 25165824  # 24 Mb
        max_original_content_length = 104857600  # 100 Mb
        streams = {}
        while not self._config.queue.empty():
            stream_name, event = await self._config.queue.get()
            if streams.get(stream_name, None) is None:
                streams[stream_name] = list()
            streams[stream_name].append(event)

        for stream, events in streams.items():
            basket = []
            basket_size = 0
            for event in events:
                if basket_size + len(event) > max_content_length:
                    count = len(basket)
                    data = self._transform_events_to_bytes(basket)
                    async with self._lock_for_count:
                        self._loop.create_task(self.send_events(stream, data, count, basket_size))
                        self.acquire_event()
                    basket.clear()
                    basket_size = 0
                basket.append(event)
                basket_size += len(event)

            data = self._transform_events_to_bytes(basket)
            async with self._lock_for_count:
                self._loop.create_task(self.send_events(stream, data, len(basket), basket_size))
                self.acquire_event()

    def _transform_events_to_bytes(self, events):
        count = len(events)
        data = count.to_bytes(4, byteorder='big') + b"".join(events)
        return data

    def force_flush(self):
        self._waiter.set()

    def close(self):
        self.running = False
        self.force_flush()

    async def send_events(self, stream, data, count_of_events, origin_size):
        self.logger.debug("Sender[send_events]. Sending events for stream {}".format(stream))
        headers = {"apiKey": self._config.api_key}
        timeout = aiohttp.ClientTimeout(
            total=self._config.send_timeout)
        
        if stream == "no_send":
            self.logger.info("Sender[send_events]. {} events NOT sended for stream {}".format(
                                    count_of_events, stream))
            async with self._lock_for_count:
                self.release_event()
            return

        if self._config.compress:
            data = lz4.frame.compress(data)
            headers["Content-Encoding"] = "lz4"
            headers["Original-Content-Length"] = origin_size

        for tries in range(3):
            host = choice(self._config.hercules_topology)
            url = "{host}/stream/send?stream={stream}".format(host=host,
                                                              stream=stream)
            self.logger.debug("Sender[send_events]. URL: {}".format(url))
            async with aiohttp.ClientSession(connector=self._tcp_connector,
                                             timeout=timeout,
                                             connector_owner=False) as session:
                try:
                    async with session.post(url, data=data, headers=headers) as resp:
                        if resp.status == 200:
                            async with self._lock_for_count:
                                self.release_event()
                            self.logger.info(
                                "Sender[send_events]. {} events send for stream {}".format(
                                    count_of_events, stream))
                            break
                    self.logger.debug(
                        "Sender[send_events]. Cannot send {} events for stream {}. Code: {}".format(
                            count_of_events, stream, resp.status))
                except asyncio.TimeoutError:
                    self.logger.debug(
                        "Sender[send_events]. Cannot send {} events for stream {}. Timeout".format(
                            count_of_events, stream))
                await asyncio.sleep(1)
        else:
            async with self._lock_for_count:
                self.release_event()
            self.logger.error(
                "Sender[send_events]. {} events for stream {} dropped after 3 tries.".format(
                    count_of_events, stream))

    def acquire_event(self):
        self._count_of_events += 1

    def release_event(self):
        self._count_of_events -= 1

    async def wait_closed(self):
        await asyncio.sleep(0.2)  # Hack. For flush
        while True:
            if self.running:
                await asyncio.sleep(0.5)
                continue
            if not self._config.queue.empty():
                await asyncio.sleep(0.5)
                continue
            async with self._lock_for_count:
                count = self._count_of_events
            if count > 0:
                await asyncio.sleep(0.5)
                continue
            break


def handler(loop, server, hercules_sender):
    loop.remove_signal_handler(signal.SIGTERM)
    loop.remove_signal_handler(signal.SIGINT)
    loop.create_task(stop_loop(server, hercules_sender))


async def stop_loop(server, hercules_sender):
    loop = asyncio.get_event_loop()
    server.close()
    await server.wait_closed()
    hercules_sender.close()
    await hercules_sender.wait_closed()
    loop.stop()


def main():
    uvloop.install()
    loop = asyncio.get_event_loop()

    try:
        config = Config(loop)
    except ValueError:
        sys.exit(1)
    except IOError:
        sys.exit(1)

    if config.tcp_message_format == "lles":
        reader = lles_reader
    if config.tcp_message_format == "lels":
        reader = lels_reader

    tcp_server_runner = asyncio.start_server(reader, config.listen, config.listen_port,
                                             limit=config.reader_buffer)
    tcp_server = loop.run_until_complete(tcp_server_runner)
    hercules_sender = HerculesSender(loop=loop, config=config)
    loop.add_signal_handler(signal.SIGTERM, handler, loop, tcp_server,
                            hercules_sender)
    loop.add_signal_handler(signal.SIGINT, handler, loop, tcp_server,
                            hercules_sender)
    try:
        loop.run_forever()
    finally:
        loop.close()


if __name__ == "__main__":
    main()
